--[[
Callback events  

TEAM_UPDATED 
	args: teamIndex or nil
		nil indicates all teams should be refreshed
		teamindex indicates which team should refresh
		
TEAM_DELETED
	args: teamIndex
		indicates which team was deleted
		
TEAM_CREATED
	args: teamIndex
		indicates the newly created teams index

SELECTED_TEAM_CHANGED
	args: teamIndex
		indicates when a new team is selected

]]


local PetBattleTeams = LibStub("AceAddon-3.0"):GetAddon("PetBattleTeams")
 TeamManager = PetBattleTeams:NewModule("TeamManager")
local PETS_PER_TEAM = 3
local START = 1
local UPDATE_PETJOURNAL = 4
local DESUMMON_PET = 5
local FINISHED = 6
local EMPTY_PET = "0x0000000000000000"

local function OnUpdate(self,elapsed)
	local selected = TeamManager:GetSelected()

	if self.step <= PETS_PER_TEAM  then
		
		local petID, abilities = TeamManager:GetPetInfo(selected,self.step)
		local isValidPet = C_PetJournal.GetPetInfoByPetID(petID or EMPTY_PET)
		local currentPetID, ability1, ability2, ability3, locked = C_PetJournal.GetPetLoadOutInfo(self.step)
		
		if not isValidPet or locked then
			C_PetJournal.SetPetLoadOutInfo(self.step,EMPTY_PET)
			if C_PetJournal.GetPetLoadOutInfo(self.step) == nil then
				self.step = self.step +1
				return
			end
		end
		
		if petID ~= currentPetID then 
			C_PetJournal.SetPetLoadOutInfo(self.step,petID) 
			return
		else
			if ability1 ~= abilities[1] then 
				C_PetJournal.SetAbility(self.step, 1, abilities[1])
				return
			elseif ability2 ~= abilities[2] then
				C_PetJournal.SetAbility(self.step, 2, abilities[2])
				return
			elseif ability3 ~= abilities[3] then
				C_PetJournal.SetAbility(self.step, 3, abilities[3])
				return
			end
		end
		
		if petID == currentPetID and ability1 == abilities[1] and ability2 == abilities[2] and ability3 == abilities[3] then
			self.step = self.step +1
			return
		end
	elseif self.step == UPDATE_PETJOURNAL then
		PetJournal_UpdatePetLoadOut()
		self.step = self.step +1
	elseif self.step == DESUMMON_PET then
		
		self.waitForPetElapsed = self.waitForPetElapsed + elapsed
		self.waitForPetTotal = self.waitForPetTotal + elapsed
		
		if self.waitForPetElapsed > 0.5 then
			if not TeamManager:GetDismissPet() then
				self.step = self.step + 1
				return
			end
			
			local summonedGuid = C_PetJournal.GetSummonedPetGUID()
			if summonedGuid then
				C_PetJournal.SummonPetByGUID(summonedGuid)
			end
			self.waitForPetElapsed = 0
		end
		
		if self.waitForPetTotal > 6 then
			self.step = self.step +1
		end
	elseif self.step == FINISHED then
		self:SetScript("OnUpdate",nil)
	end 
end

function TeamManager:GetPetInfo(teamIndex,petIndex)
	assert(type(teamIndex) == "number" and type(petIndex) == "number")
	if self.teams[teamIndex] and self.teams[teamIndex][petIndex] then
		local pet = self.teams[teamIndex][petIndex]
		if pet and C_PetJournal.GetPetInfoByPetID(pet.petID or EMPTY_PET) then
			local abilities = pet.abilities
			return pet.petID, abilities
		end
	end
	return EMPTY_PET
end

function TeamManager:GetNumPets(teamIndex)
	assert(type(teamIndex) == "number")
	if self.teams[teamIndex] then 
		return #self.teams[teamIndex]
	end
	return 0
end

function TeamManager:GetNumTeams()
	return #self.teams
end

function TeamManager:GetTeamName(teamIndex)
	if type(teamIndex) ~= "number" then return "Unknown Team" end
	
	local name = "Team: "..tostring(teamIndex)
	local customName
	
	if self.teams[teamIndex] and self.teams[teamIndex].name then
		customName = self.teams[teamIndex].name
	end
	
	local displayName = customName and customName or name
	
	return displayName, name, customName
end

function TeamManager:SetTeamName(teamIndex,name)
	assert(type(teamIndex) == "number"  and  (type(name) == "string" or name == nil))
	if  self.teams[teamIndex] then
		self.teams[teamIndex].name = name
	end
	self.callbacks:Fire("TEAM_UPDATED",teamIndex)
end

function TeamManager:MoveTeam(teamIndexSource,teamIndexDesintation)
	assert(type(teamIndexSource) == "number" and type(teamIndexDesintation) == "number")
	if self.teams[teamIndexSource] and self.teams[teamIndexDesintation] then
		
		local selected = self:GetSelected()
		local distance = math.abs(teamIndexSource - teamIndexDesintation)
		
		if true then --distance > 1 then
			local tempTeam = self.teams[teamIndexSource] 
			table.remove(self.teams,teamIndexSource)
			
			 --adjust for the fact that we delete first to make sure their is consistant behavior for the user
			if teamIndexSource < teamIndexDesintation then
				teamIndexDesintation = teamIndexDesintation - 1
			end
			
			table.insert(self.teams,teamIndexDesintation ,tempTeam )
			
			--if the swap happens in such as way as to re-order the teams pushing the selected team to another team number
			--then update the selected team to the predicted new location of the old selected team
			
			--if source && desintation < selected or source && desintation > selected do nothing
			--if destination == selected and source < selected then post move do nothing
			
			--if source < selected and destination > selected then post move, selected is -1
			--if source == selected then post move selected is desintation - 1
			
			--if destination == selected and source > selected then post move selected +1
			--if source > selected and destination < selected then post move, selected is +1
			
			local adjust = teamIndexSource <  teamIndexDesintation and -1 or 1
			
			if teamIndexSource == selected then
				self:SetSelected(teamIndexDesintation)
			elseif teamIndexDesintation == selected   then
				self:SetSelected(teamIndexDesintation + adjust)
			elseif (teamIndexSource < selected and teamIndexDesintation > selected) or (teamIndexSource > selected and teamIndexDesintation < selected) then
				self:SetSelected(selected + adjust)
			end
			
			self.callbacks:Fire("TEAM_UPDATED")
		--[[else 
			local tempTeam = self.teams[teamIndexSource] 
			self.teams[teamIndexSource]  = self.teams[teamIndexDesintation]
			self.teams[teamIndexDesintation] = tempTeam
			
			--swap selection
			if selected == teamIndexSource then
				self:SetSelected(teamIndexDesintation)
			elseif selected == teamIndexDesintation then
				self:SetSelected(teamIndexSource)
			end

			self.callbacks:Fire("TEAM_UPDATED")]]
		end
	end
end

function TeamManager:CreateTeam()
	local numTeams, maxTeams = self:GetNumTeams()
	
	local team = {}
	team.name = nil;
	
	
		for i = 1,PETS_PER_TEAM do
			pet = {}
			pet.abilities = {}
			
			if self:GetAutomaticallySaveTeams() then 
				pet.petID, pet.abilities[1], pet.abilities[2], pet.abilities[3] = C_PetJournal.GetPetLoadOutInfo(i)
			else
				pet.petID, pet.abilities[1], pet.abilities[2], pet.abilities[3] = EMPTY_PET,0,0,0
			end
	
			table.insert(team,pet)
		end
	
	table.insert(self.teams,self:GetSelected()+1,team)
	self.callbacks:Fire("TEAM_CREATED",self:GetSelected()+1)
	self.callbacks:Fire("TEAM_UPDATED",self:GetSelected()+1)
	self:SetSelected(self:GetSelected()+1)
end


function TeamManager:TeamExists(teamIndex)
	assert(type(teamIndex) == "number")
	if self.teams[teamIndex] then return true end
	return false
end

function TeamManager:TeamContainsPet(teamIndex,petID)
	assert(type(teamIndex) == "number")
	
	if not self.teams[teamIndex] or not petID or petID == EMPTY_PET then return false end
	local numPets = TeamManager:GetNumPets(teamIndex)
	for i=1,numPets do
		local teamPetID = TeamManager:GetPetInfo(teamIndex,i)
		if teamPetID == petID then
			return true,i
		end
	end
	return false
end

function TeamManager:DeleteTeam(teamIndex)
	assert(type(teamIndex) == "number")
	
	
	local numTeams = self:GetNumTeams()
	local selectedTeam = self:GetSelected()
	
	if numTeams > 1 then
		table.remove(self.teams,teamIndex)
		
		if teamIndex == numTeams then
			self:SetSelected(teamIndex-1)
		elseif teamIndex == selectedTeam then
			self:SetSelected(teamIndex)
		end
		
		self.callbacks:Fire("TEAM_DELETED",teamIndex)
	end
end

function TeamManager:IsWorking()
	return self.frame.step < DESUMMON_PET
end

function TeamManager.UpdateCurrentTeam()
	if not TeamManager:GetAutomaticallySaveTeams() then return end
	
	local selected = TeamManager:GetSelected()
	
	if not TeamManager:IsWorking() and not TeamManager:IsTeamLocked(selected) and not TeamManager:IsTeamLockedByUser(selected) then
		
		local team =  TeamManager.teams[selected]
		for i = 1,PETS_PER_TEAM do
			pet = {}
			pet.abilities = {}
			pet.petID, pet.abilities[1], pet.abilities[2], pet.abilities[3] = C_PetJournal.GetPetLoadOutInfo(i)
			team[i] = pet
		end
		
		TeamManager.callbacks:Fire("TEAM_UPDATED",selected)
	end
end

function TeamManager:GetSelected()
	return self.db.global.selected 
end

function TeamManager:GetShowTeamName()
	return self.db.global.showTeamName 
end
function TeamManager:GetShowXpInLevel()
	return self.db.global.showXpInLevel 
end
function TeamManager:GetShowXpInHealthBar()
	return self.db.global.showXpInHealthBar 
end

function TeamManager:SetShowTeamName(enabled)
	assert(type(enabled) == "boolean")
	self.db.global.showTeamName = enabled
	self.callbacks:Fire("TEAM_UPDATED")
end

function TeamManager:SetShowXpInLevel(enabled)
	assert(type(enabled) == "boolean")
	self.db.global.showXpInLevel = enabled
	self.callbacks:Fire("TEAM_UPDATED")
end
function TeamManager:SetShowXpInHealthBar(enabled)
	assert(type(enabled) == "boolean")
	self.db.global.showXpInHealthBar = enabled 
	self.callbacks:Fire("TEAM_UPDATED")
end

function TeamManager:IsSelected(teamIndex)
	assert(type(teamIndex) == "number")
	return self.db.global.selected  == teamIndex
end 


function TeamManager:SetSelected(teamIndex)
	assert(type(teamIndex) == "number" )
	local selected = self:GetSelected()
	if teamIndex <= 0 or teamIndex > self:GetNumTeams() or self:IsTeamLocked(selected) then return end
	local prevSelected = self.db.global.selected
	self.db.global.selected = teamIndex
	TeamManager.callbacks:Fire("TEAM_UPDATED",prevSelected)
	self.callbacks:Fire("SELECTED_TEAM_CHANGED",teamIndex)
	self:ApplyTeam(teamIndex)
end

function TeamManager:IsTeamLocked(teamIndex)
	assert(type(teamIndex) == "number")
	if teamIndex == self:GetSelected() and (C_PetBattles.GetPVPMatchmakingInfo()) then return true end
	return false
end

function TeamManager:IsTeamLockedByUser(teamIndex)
	assert(type(teamIndex) == "number")
	if self.teams[teamIndex] and self.teams[teamIndex].locked == true then return true end
	return false
end

function TeamManager:UpdateTeamNewPet(petID,teamIndex,petIndex,sourceTeamIndex,sourcePetIndex)
	assert(type(petID) == "string")
	assert(type(petIndex) == "number" and petIndex <= PETS_PER_TEAM and petIndex > 0)
	assert(type(teamIndex) == "number")

	if self:IsTeamLocked(teamIndex) or self:IsTeamLockedByUser(teamIndex) then return end
	if self:TeamContainsPet(teamIndex,petID) then return  end
	local speciesID, customName, level, xp, maxXp, displayID, _,name, icon, petType, creatureID, sourceText, description, isWild, canBattle, tradable = C_PetJournal.GetPetInfoByPetID(petID)
	
	if speciesID and canBattle then
		local pet = {}
		pet.petID = petID
		pet.abilities = {}
		if sourceTeamIndex and sourcePetIndex and  sourcePetIndex < PETS_PER_TEAM then
			local _, sourceAbilities = self:GetPetInfo(sourceTeamIndex,sourcePetIndex)
			
			for i=1,3 do
				pet.abilities[i] = sourceAbilities[i]
			end
			
		else
			local abilities = {}
			C_PetJournal.GetPetAbilityList(speciesID, abilities, {})
			for i=1,3 do
				pet.abilities[i] = abilities[i]
			end
		end
			
		self.teams[teamIndex][petIndex] = pet;
		if self:IsSelected(teamIndex) then
			self:ApplyTeam(teamIndex)
		end
		self.callbacks:Fire("TEAM_UPDATED",teamIndex)
		return 
	end
	return
end

function TeamManager:UpdateTeamSwapPets(destinationTeam,destinationPetIndex,sourceTeam,sourcePetIndex)
	assert(type(destinationTeam) == "number")
	assert(type(destinationPetIndex) == "number" and destinationPetIndex <= PETS_PER_TEAM and destinationPetIndex > 0)
	assert(type(sourceTeam) == "number")
	assert(type(sourcePetIndex) == "number" and sourcePetIndex <= PETS_PER_TEAM and sourcePetIndex > 0)
	
	
	
	local selected = self:GetSelected()
	if self:IsTeamLocked(destinationTeam) or self:IsTeamLocked(sourceTeam) then return end
	if self:IsTeamLockedByUser(destinationTeam) or self:IsTeamLockedByUser(sourceTeam) then return end
	
	local petA = self.teams[destinationTeam][destinationPetIndex]
	local petB = self.teams[sourceTeam][sourcePetIndex]
	
	if destinationTeam ~=  sourceTeam and  ( (petB and self:TeamContainsPet(destinationTeam,petB.petID)) or (petA and self:TeamContainsPet(sourceTeam,petA.petID))) then return end
	
	self.teams[destinationTeam][destinationPetIndex] = petB
	self.teams[sourceTeam][sourcePetIndex] = petA
	
	if selected == sourceTeam or selected == destinationTeam then 
		self:ApplyTeam(selected)
	end
	
	
	self.callbacks:Fire("TEAM_UPDATED",destinationTeam)
	self.callbacks:Fire("TEAM_UPDATED",sourceTeam)
end

function TeamManager:OnInitialize()
	self.callbacks = LibStub("CallbackHandler-1.0"):New(self)
	self.frame = CreateFrame("frame")
	self.frame.step = FINISHED
	
	local defaults = {
		global = {
			teams = {},
			selected = 0,
			dismissPet = false,
			hasImported = false,
			userLocked = false,
			showTeamName = true,
			showXpInLevel = true,
			showXpInHealthBar = false,
			automaticallySaveTeams = true,
		}
	}

	-- Assuming the .toc says ## SavedVariables: MyAddonDB
	local db = LibStub("AceDB-3.0"):New("PetBattleTeamsDB", {} , true)
	local name = self:GetName()
	self.db = db:RegisterNamespace(name, defaults)
	
	self.teams = self.db.global.teams 
	
	
	if self.db.global.hasImported == false then
		StaticPopup_Show("PBT_IMPORT_TEAMS",nil,nil,self)
		self.db.global.hasImported = true
	end
	
	--convert numbered petID's to new hex strings
	for i=1,#self.teams do
		for j=1,3 do
			if self.teams[i][j] and self.teams[i][j].petID and type(self.teams[i][j].petID) == "number" then
				self.teams[i][j].petID = string.format("%0#18x",  self.teams[i][j].petID)
			end
		end
	end
	
	if #self.teams == 0 then
		self:CreateTeam()
		self:SetSelected(1)
	end
	
	if #self.teams ~= 0 and self:GetSelected() == 0 then
		self:SetSelected(1)
	end
	 
	hooksecurefunc(C_PetJournal, "SetPetLoadOutInfo", TeamManager.UpdateCurrentTeam)
	hooksecurefunc(C_PetJournal, "SetAbility", TeamManager.UpdateCurrentTeam)
end


function TeamManager:ApplyTeam(teamIndex)
	assert(type(teamIndex) == "number")
	self.frame.step = START
	self.frame.waitForPetElapsed = 1 --so it fires immediately
	self.frame.waitForPetTotal = 0
	self.frame:SetScript("OnUpdate",OnUpdate)
end

--/run TeamManager:ImportTeams(PetBattleTeamsSettings.teams)
function TeamManager:ImportTeams()
	if PetBattleTeamsSettings and PetBattleTeamsSettings.teams and #PetBattleTeamsSettings.teams > 0 then
		local teams = PetBattleTeamsSettings.teams
		local start = self:GetNumTeams() +1
		for i=1,#teams do
			local team = {}
			for j=1,PETS_PER_TEAM do
				local pet = {}
				pet.abilities = {}
				pet.petID = teams[i][j].petID
				for k=1,#teams[i][j].abilities do
					pet.abilities[k] = teams[i][j].abilities[k]
				end
				table.insert(team,pet)
			end
			table.insert(self.teams,team)
		end
		self:SetSelected(1)
		self.callbacks:Fire("TEAM_UPDATED")
		print("PetBattle Teams: Imported",#teams,"Teams.")
	end
end

function TeamManager:ResetTeams()
	wipe(self.teams)
	self:CreateTeam()
	print("PetBattle Teams: Teams Reset")
end

function TeamManager:SetLockStateAllTeams(enabled)
	assert(type(enabled)== "boolean")
	for i=1,#self.teams do
		self.teams[i].locked = enabled
	end
	self.callbacks:Fire("TEAM_UPDATED")
	print("PetBattle Teams: Teams Unlocked")
end

function TeamManager:LockTeam(teamIndex,enabled)
	assert(type(enabled)== "boolean" and type(teamIndex) == "number")
	if self.teams[teamIndex] then
		self.teams[teamIndex].locked = enabled
	end
	self.callbacks:Fire("TEAM_UPDATED",teamIndex)
end

function TeamManager:SetAutomaticallySaveTeams(enabled)
	assert(type(enabled)== "boolean")
	self.db.global.automaticallySaveTeams = enabled
end

function TeamManager:GetAutomaticallySaveTeams()
	return self.db.global.automaticallySaveTeams
end

function TeamManager:SetDismissPet(enabled)
	assert(type(enabled)== "boolean")
	self.db.global.dismissPet = enabled
end
	
function TeamManager:GetDismissPet()
	return self.db.global.dismissPet
end

function TeamManager:ResetUI()
	self:SetShowTeamName(true)
	self:SetShowXpInLevel(true)
	self:SetShowXpInHealthBar(false)
	self:SetAutomaticallySaveTeams(true)
	self:SetLockStateAllTeams(false)
	self:SetSelected(1)
end











